import os
import re
from typing import Optional

import yaml

from ..constants import PBT_FILE_NAME, LANGUAGE, JOBS, PIPELINES, \
    PIPELINE_CONFIGURATIONS, CONFIGURATIONS, JSON_EXTENSION, BASE_PIPELINE, PROJECT_ID_PLACEHOLDER_REGEX, \
    PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX, PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX, GEMS
from ..exceptions import ProjectPathNotFoundException, ProjectFileNotFoundException

SUBSCRIBED_ENTITY_URI_REGEX = r"gitUri=(.*)&subPath=(.*)&tag=(.*)&projectSubscriptionProjectId=(.*)&path=(.*)"


class Project:
    _DATABRICKS_JOB_JSON = "databricks-job.json"

    def __init__(self, project_path: str, project_id: Optional[str] = None,
                 release_tag: Optional[str] = None, release_version: Optional[str] = None):

        self.project_id = project_id
        self.project_path = project_path

        self.release_tag = release_tag
        self.release_version = release_version

        self.pbt_project_dict = {}
        self.project_language = None
        self.jobs = None
        self.pipelines = None
        self.pipeline_id_to_name = {}
        self.gems = {}

        self._verify_project()
        self._load_project_config()
        self._extract_project_info()
        self.pipeline_configurations = self._load_pipeline_configurations()

    def load_databricks_job(self, job_id: str) -> Optional[str]:
        try:
            path = os.path.join(self.project_path, job_id, "code", self._DATABRICKS_JOB_JSON)
            content = self._read_file_content(path)

            if content is not None:
                return self._replace_placeholders(self._DATABRICKS_JOB_JSON, content)

            return content
        except Exception:
            return None

    def load_airflow_base_folder_path(self, job_id):
        return os.path.join(self.project_path, job_id, "code")

    def load_airflow_folder(self, job_id):
        return self._read_directory(os.path.join(self.project_path, job_id, "code"))

    def load_airflow_folder_with_placeholder(self, job_id):
        return self._read_directory_with_placeholder(os.path.join(self.project_path, job_id, "code"))

    def load_airflow_aspect(self, job_id: str) -> Optional[str]:
        path = os.path.join(self.project_path, job_id, "pbt_aspects.yml")
        return self._read_file_content(path)

    def load_pipeline_base_path(self, pipeline):
        return os.path.join(self.project_path, pipeline, "code")

    @staticmethod
    def is_cross_project_pipeline(pipeline):

        match = re.compile(SUBSCRIBED_ENTITY_URI_REGEX).search(pipeline)

        if match:
            git_uri, sub_path, tag, project_subscription_project_id, path = match.groups()
            return project_subscription_project_id, tag.split('/')[1], path
        else:
            return None, None, None

    def load_pipeline_folder(self, pipeline):
        (project_subscription_id, version, pipeline_path) = self.is_cross_project_pipeline(pipeline)

        if project_subscription_id is not None:
            sub_path = f".prophecy/{project_subscription_id}/{pipeline_path}"
        else:
            sub_path = pipeline

        base_path = os.path.join(self.project_path, sub_path, "code")

        return self._read_directory(base_path)

    @staticmethod
    def get_pipeline_whl_or_jar(base_path: str):
        for dir_path, dir_names, filenames in os.walk(base_path):
            for filename in filenames:
                full_path = os.path.join(dir_path, filename)
                if full_path.endswith(".whl") or full_path.endswith(".jar"):
                    return full_path

    def get_py_pipeline_main_file(self, pipeline_id):
        subscribed_project_id, release_tag, pipeline_path = self.is_cross_project_pipeline(pipeline_id)

        if subscribed_project_id is None:
            path = pipeline_id
        else:
            path = ".prophecy/{}/{}".format(subscribed_project_id, pipeline_path)

        main_file = os.path.join(self.project_path, path, "code", "main.py")
        data = self._read_file_content(main_file)

        return data

    def get_pipeline_name(self, pipeline_id):
        subscribed_project_id, release_tag, pipeline_path = self.is_cross_project_pipeline(pipeline_id)

        # pipeline belongs to same deployment.
        if subscribed_project_id is None:
            pipeline_path = pipeline_id
            pipelines = self.pipelines

        else:
            subscribed_project = self._load_subscribed_project_yml(subscribed_project_id)
            pipelines = subscribed_project.get('pipelines', {})

        return pipelines.get(pipeline_path, {}).get('name', pipeline_path.split('/')[-1])

    def _read_directory_with_placeholder(self, base_path: str):
        rdc = {}

        for dir_path, dir_names, filenames in os.walk(base_path):
            for filename in filenames:
                if filename.endswith(('.py', '.json', '.conf', '.scala', 'pom.xml')):
                    full_path = os.path.join(dir_path, filename)
                    content = self._read_file_content(full_path)
                    if content is not None:
                        relative_path = os.path.relpath(full_path, base_path)
                        rdc[relative_path] = content
        return rdc

    def _read_directory(self, base_path: str):
        rdc = {}

        for dir_path, dir_names, filenames in os.walk(base_path):
            for filename in filenames:
                if filename.endswith(('.py', '.json', '.conf', '.scala', 'pom.xml')):
                    full_path = os.path.join(dir_path, filename)
                    content = self._read_file_content(full_path)
                    if content is not None:
                        relative_path = os.path.relpath(full_path, base_path)
                        rdc[relative_path] = self._replace_placeholders(relative_path, content)
        return rdc

    def _load_subscribed_project_yml(self, subscribed_project_id: str):
        path = os.path.join(self.project_path, ".prophecy", subscribed_project_id, "pbt_project.yml")
        content = self._read_file_content(path)
        if content is not None:
            return yaml.safe_load(content)

    def _verify_project(self):
        if not os.path.exists(self.project_path):
            raise ProjectPathNotFoundException(f"Project path does not exist {self.project_path}")

        if not os.path.exists(os.path.join(self.project_path, PBT_FILE_NAME)):
            raise ProjectFileNotFoundException(f"Project file does not exist at path {self.project_path}")

    def _load_project_config(self):
        pbt_project_path = os.path.join(self.project_path, PBT_FILE_NAME)
        content = self._read_file_content(pbt_project_path)
        if content is not None:
            self.pbt_project_dict = yaml.safe_load(content)

    def _extract_project_info(self):
        self.project_language = self.pbt_project_dict.get(LANGUAGE, None)
        self.jobs = self.pbt_project_dict.get(JOBS, {})
        self.pipelines = self.pbt_project_dict.get(PIPELINES, {})
        self.gems = self.pbt_project_dict.get(GEMS, {})

    def _load_pipeline_configurations(self):
        pipeline_configurations = {}

        pipeline_conf = dict(self.pbt_project_dict.get(PIPELINE_CONFIGURATIONS, []))

        for pipeline_config_path, pipeline_config_object in pipeline_conf.items():
            configurations = {}

            for configurations_key in pipeline_config_object[CONFIGURATIONS]:
                config_name = pipeline_config_object[CONFIGURATIONS][configurations_key]["name"]
                file_path = os.path.join(self.project_path, configurations_key + JSON_EXTENSION)

                configurations[config_name] = self._read_file_content(file_path)
            pipeline_configurations[pipeline_config_object[BASE_PIPELINE]] = configurations

        return pipeline_configurations

    def _replace_placeholders(self, path: str, content: str) -> str:
        if path.endswith('.json') or path.endswith('.scala') or path.endswith('.py'):
            content = content.replace(PROJECT_ID_PLACEHOLDER_REGEX, self.project_id) \
                .replace(PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX, self.release_version) \
                .replace(PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX, self.release_tag)

        return content

    def _read_file_content(self, file_path: str) -> Optional[str]:
        content = None
        with open(file_path, 'r') as file:
            content = file.read()
        return content

    # only check for non-empty gems directory
    def non_empty_gems_directory(self):
        return os.path.exists(os.path.join(self.project_path, "gems"))



from typing import Optional

from .deployment.project import ProjectDeployment
from .entities.project import Project
from .project_config import ProjectConfig


class PBTCli(object):
    """Command line interface for PBT."""

    def __init__(self, project: Project, project_config: ProjectConfig):
        self.project = ProjectDeployment(project, project_config)

    def headers(self):
        """Print headers."""
        self.project.headers()

    def build(self):
        """Build pipelines."""
        self.project.build()

    def deploy(self, job_ids):
        """Deploy pipelines."""
        self.project.deploy(job_ids)

    @classmethod
    def from_conf_folder(cls, project_path: str, project_id: str, conf_folder: str, release_tag: Optional[str],
                         release_version: str):
        """Create PBTCli from conf folder."""
        project = Project(project_path, project_id, release_tag, release_version)
        project_config = ProjectConfig.from_conf_folder(conf_folder)
        return cls(project, project_config)



# back-off :- 0 for making this fixed delay based retry
import base64
import binascii
import hashlib
import os
from enum import Enum
from typing import Optional, Any

from .project_models import LogEvent, StepMetadata, Status


def calculate_checksum(input_str, salt=None):
    salt = salt or "better_salt_than_never"

    md = hashlib.sha256()
    input_bytes = input_str.encode('utf-8')
    salt_bytes = salt.encode('utf-8')

    md.update(input_bytes)
    md.update(salt_bytes)

    digest_bytes = md.digest()

    return ''.join(f'{byte:02x}' for byte in digest_bytes)


def generate_secure_content(content: str, salt: str) -> str:
    iterations = 10000
    key_length = 128

    password = content.encode('utf-8')
    salt_bytes = salt.encode('utf-8')

    derived_key = hashlib.pbkdf2_hmac('sha256', password, salt_bytes, iterations, dklen=key_length // 8)
    hash_bytes = binascii.hexlify(derived_key)

    return base64.b64encode(hash_bytes).decode('utf-8').replace('\\W+', '_')


# class mimicking Either behaviour to capture both success and failure
class Either:

    def __init__(self, left=None, right=None):
        self.left = left
        self.right = right

    @property
    def is_left(self):
        return self.left is not None

    @property
    def is_right(self) -> bool:
        return self.right is not None


def custom_print(message: Optional[Any] = None, exception=None,
                 step_id=None,
                 step_metadata: Optional[StepMetadata] = None,
                 step_status: Optional[Status] = None):
    if os.environ.get('PRINT_MODE', 'REGULAR') == 'CUSTOM':
        # Custom print: Print all variables.
        if step_metadata is not None:
            log_event = LogEvent.from_step_metadata(step_metadata)
        elif step_status is not None:
            log_event = LogEvent.from_status(step_status, step_id)
        else:
            log_event = LogEvent.from_log(step_id, message, exception)

        print(log_event.to_json(), flush=True)
    else:
        # Regular print: Skip stepName.
        print(message, exception)


# If the item is a dictionary


def remove_null_items_recursively(item):
    # If the item is a dictionary
    if isinstance(item, dict):
        return {
            k: remove_null_items_recursively(v)
            for k, v in item.items() if v is not None
        }

    # If the item is a list
    elif isinstance(item, list):
        return [remove_null_items_recursively(v) for v in item if v is not None]

    # If the item is an enum
    elif isinstance(item, Enum):
        return item.value

    # If the item is neither a dictionary, a list, nor an enum
    else:
        return item



import enum
import os
from typing import List, Optional

from pydantic import BaseModel
from pydantic_yaml import parse_yaml_raw_as

from .constants import PROPHECY_ARTIFACTS, DBFS_FILE_STORE
from .deployment import OperationType, JobInfoAndOperation
from .utility import Either


class SchedulerType(enum.Enum):
    Composer = "Composer"
    MWAA = "MWAA"
    Databricks = "Databricks"
    Prophecy = "Prophecy"
    EMR = "EMR"

    @staticmethod
    def from_fabric_provider(provider_type: str):
        # Convert the provider_type to its equivalent Enum if exists, otherwise None
        return SchedulerType[provider_type]


class FabricType(enum.Enum):
    Spark = "Spark"
    Sql = "Sql"
    Airflow = "Airflow"


class FabricProviderType(enum.Enum):
    Composer = "Composer"
    MWAA = "MWAA"
    Databricks = "Databricks"
    Prophecy = "Prophecy"
    EMR = "EMR"
    Dataproc = "Dataproc"


class RuntimeMode(enum.Enum):
    Regular = "Regular"
    Partial = "Partial"
    RegularWithTests = "RegularWithTests"
    PartialWithTests = "PartialWithTests"
    Test = "Test"


class DeploymentMode(enum.Enum):
    FullProject = "FullProject"
    SelectiveJob = "SelectiveJob"


class EMRInfo(BaseModel):
    region: str
    bucket: str
    access_key_id: str
    secret_access_key: str
    session_token: Optional[str] = None

    def _bucket(self):
        return self.bucket.replace("s3://", "").split("/")

    def bare_bucket(self):
        return self._bucket()[0]

    def bare_path_prefix(self):
        return "/".join(self._bucket()[1:])


class DataprocInfo(BaseModel):
    bucket: str
    project_id: str
    key_json: str
    location: str

    def _bucket(self):
        return self.bucket.replace("gs://", "").split("/")

    def bare_bucket(self):
        return self._bucket()[0]

    def bare_path_prefix(self):
        return "/".join(self._bucket()[1:])


class ComposerInfo(BaseModel):
    key_json: str
    version: str
    location: Optional[str] = None
    project_id: str
    airflow_url: str
    dag_location: str
    client_id: Optional[str] = None


class MwaaInfo(BaseModel):
    region: str
    version: str
    access_key: str
    secret_key: str
    airflow_url: str
    dag_location: str
    environment_name: str


class DatabricksInfo(BaseModel):
    url: str
    token: str


class FabricInfo(BaseModel):
    id: str
    name: str
    type: Optional[FabricType]  # sql/ databricks/ airflow
    provider: Optional[FabricProviderType]  # composer/mwaa/prophecy/databricks/dataproc
    composer: Optional[ComposerInfo] = None
    mwaa: Optional[MwaaInfo] = None
    databricks: Optional[DatabricksInfo] = None
    emr: Optional[EMRInfo] = None
    dataproc: Optional[DataprocInfo] = None


class JobInfo(BaseModel):
    name: str
    type: SchedulerType
    external_job_id: str
    fabric_id: str
    id: str
    is_paused: Optional[bool] = False
    skip_processing: Optional[bool] = False  # this is useful in case when we deploy from older release tags.
    release_tag: Optional[str] = None

    def with_release_tag(self, release_tag):
        self.release_tag = release_tag

    @staticmethod
    def create_db_job(name: str, id: str, fabric_id: str, external_job_id: str, release_tag: str,
                      is_paused: bool = False):
        return JobInfo(name=name, type=SchedulerType.Databricks, id=id, fabric_id=fabric_id,
                       external_job_id=external_job_id,
                       release_tag=release_tag, is_paused=is_paused)

    @staticmethod
    def create_airflow_job(name: str, id: str, fabric_id: str, external_job_id: str, release_tag: str,
                           is_paused: bool = False, fabric_provider_type: str = ""):
        return JobInfo(name=name, type=SchedulerType.from_fabric_provider(fabric_provider_type), id=id,
                       fabric_id=fabric_id,
                       external_job_id=external_job_id,
                       release_tag=release_tag, is_paused=is_paused)

    def is_exactly_same_as(self, job_response):
        return self.external_job_id == job_response.external_job_id and self.fabric_id == job_response.fabric_id and \
            self.id == job_response.id and self.name == job_response.name and self.type == job_response.type

    def pause(self, flag: bool):
        self.is_paused = flag


class ProjectAndGitTokens(BaseModel):
    project_id: str
    git_token: str = ""


class FabricConfig(BaseModel):
    fabrics: List[FabricInfo] = []
    project_git_tokens: List[ProjectAndGitTokens] = []

    def is_fabric_db_fabric(self, fabric_id: str) -> bool:
        return any((fabric for fabric in self.fabrics if
                    fabric.id == fabric_id and fabric.provider == FabricProviderType.Databricks))

    def contains_fabric(self, fabric_id: str) -> bool:
        return any(fabric.id == fabric_id for fabric in self.fabrics)

    def get_fabric(self, fabric_id: str) -> Optional[FabricInfo]:
        return next((fabric for fabric in self.fabrics if fabric.id == fabric_id), None)

    def git_token_for_project(self, project_id: str) -> Optional[str]:
        return next((project.git_token for project in self.project_git_tokens if project.project_id == project_id),
                    None)

    def is_fabric_prophecy_managed(self, fabric_id: str) -> bool:
        if fabric_id is not None and self.get_fabric(fabric_id) is not None:
            fabric_info = self.get_fabric(fabric_id)
            return fabric_info.type == FabricType.Airflow and fabric_info.provider == FabricProviderType.Prophecy

        return False

    def is_fabric_emr_fabric(self, fabric_id: str) -> bool:
        return any((fabric for fabric in self.fabrics if
                    fabric.id == fabric_id and fabric.provider == FabricProviderType.EMR))

    def db_fabrics(self):
        return [fabric.id for fabric in self.fabrics if
                (fabric.type == FabricType.Spark and fabric.provider == FabricProviderType.Databricks) or
                (fabric.type == FabricType.Sql and fabric.provider == FabricProviderType.Databricks)]

    def emr_fabrics(self) -> List[FabricInfo]:
        return [fabric for fabric in self.fabrics if
                (fabric.type == FabricType.Spark and fabric.provider == FabricProviderType.EMR)]

    def dataproc_fabrics(self) -> List[FabricInfo]:
        return [fabric for fabric in self.fabrics if
                (fabric.type == FabricType.Spark and fabric.provider == FabricProviderType.Dataproc)]

    def list_all_fabrics(self) -> List[str]:
        return [fabric.id for fabric in self.fabrics]


class JobsState(BaseModel):
    version: str
    jobs: List[JobInfo] = []

    def contains_jobs(self, job_id: str, fabric_uid: str) -> bool:
        return any(
            job.id == job_id and job.fabric_id == fabric_uid and job.skip_processing is False for job in self.jobs)

    def get_jobs(self, job_id: str) -> List[JobInfo]:
        return [job for job in self.jobs if job.id == job_id and job.skip_processing is False]

    def get_job(self, job_id: str, fabric_id: str) -> Optional[JobInfo]:
        return next(
            (job for job in self.jobs if
             job.id == job_id and job.fabric_id == fabric_id and job.skip_processing is False),
            None)

    @property
    def get_databricks_jobs(self) -> List[JobInfo]:
        return [job for job in self.jobs if job.type == SchedulerType.Databricks and job.skip_processing is False]

    @property
    def get_airflow_jobs(self) -> List[JobInfo]:
        return [job for job in self.jobs if job.type != SchedulerType.Databricks and job.skip_processing is False]

    def filter_jobs(self, job_info) -> List[JobInfo]:
        return [job for job in self.jobs if
                not (job.id == job_info.id and job.fabric_id == job_info.fabric_id)]

    def update_state(self, jobs_responses: List[Either]):
        filtered_response: List[JobInfoAndOperation] = [response.right for
                                                        response
                                                        in jobs_responses if
                                                        response.is_right]
        # Important to do all operations in this order,
        # first we delete
        # then we refresh
        # then we create

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.DELETED:
                self.jobs = self.filter_jobs(job_info)

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.REFRESH:
                new_jobs = self.filter_jobs(job_info)
                new_jobs.append(job_info)
                self.jobs = new_jobs

        for job_response in filtered_response:
            job_info = job_response.job_info

            if job_response.operation == OperationType.CREATED:
                matching_job = next((job for job in self.jobs if job.id == job_info.id), None)

                new_jobs = self.filter_jobs(job_info)

                if matching_job is not None:
                    # super important to preserve the release tag.
                    job_info.with_release_tag(matching_job.release_tag)

                new_jobs.append(job_info)

                self.jobs = new_jobs


class JobsAndFabric(BaseModel):
    job_id: str
    fabric_id: str


class ConfigsOverride(BaseModel):
    are_tests_enabled: Optional[bool] = False
    mode: Optional[DeploymentMode] = DeploymentMode.FullProject
    jobs_and_fabric: Optional[List[JobsAndFabric]] = None

    def find_fabric_override_for_job(self, job_id: str) -> Optional[str]:
        if self.mode == DeploymentMode.SelectiveJob and self.jobs_and_fabric:
            matching_fabric = next(
                (job_and_fabric.fabric_id for job_and_fabric in self.jobs_and_fabric if
                 job_and_fabric.job_id == job_id),
                None
            )
            return matching_fabric
        return None

    @staticmethod
    def empty():
        return ConfigsOverride()

    def is_job_to_run(self, job_id) -> bool:
        if self.mode == DeploymentMode.SelectiveJob and self.jobs_and_fabric:
            return any(job_and_fabric.job_id == job_id for job_and_fabric in self.jobs_and_fabric)
        return True


class NexusConfig(BaseModel):
    url: str
    username: str
    password: str
    repository: str


class SystemConfig(BaseModel):
    customer_name: Optional[str] = 'dev'
    control_plane_name: Optional[str] = 'execution'
    prophecy_salt: Optional[str] = 'execution'
    nexus: Optional[NexusConfig] = None

    def get_dbfs_base_path(self):
        return f'{DBFS_FILE_STORE}/{PROPHECY_ARTIFACTS}/{self.customer_name}/{self.control_plane_name}'

    def get_s3_base_path(self):
        return f"{PROPHECY_ARTIFACTS}/{self.customer_name}/{self.control_plane_name}"

    @staticmethod
    def empty():
        return SystemConfig()


class ProjectConfig:
    def __init__(self, jobs_state: JobsState, fabric_config: FabricConfig, system_config: SystemConfig,
                 config_override: ConfigsOverride):

        self.jobs_state = jobs_state
        self.fabric_config = fabric_config
        self.system_config = system_config
        self.configs_override = config_override

    @staticmethod
    def from_path(job_state_path: str, system_config_path: str, configs_override_path: str,
                  fabric_config_path: str) \
            :
        def load_jobs_state():
            if job_state_path is not None and len(job_state_path) > 0:
                with open(job_state_path, "r") as job_state:
                    data = job_state.read()
                    return parse_yaml_raw_as(JobsState, data)
            else:
                raise Exception("Job state config path is not provided")

        def load_system_config():
            if system_config_path is not None and len(system_config_path) > 0:
                with open(system_config_path, "r") as system_config:
                    data = system_config.read()
                    return parse_yaml_raw_as(SystemConfig, data)
            else:
                raise Exception("System config path is not provided")

        def load_configs_override():
            if configs_override_path is not None and len(configs_override_path) > 0:
                with open(configs_override_path, "r") as config_override:
                    data = config_override.read()
                    return parse_yaml_raw_as(ConfigsOverride, data)
            else:
                return ConfigsOverride.empty()

        def load_fabric_config():
            if fabric_config_path is not None and len(fabric_config_path) > 0:
                with open(fabric_config_path, "r") as fabric_config:
                    data = fabric_config.read()
                    return parse_yaml_raw_as(FabricConfig, data)
            else:
                raise Exception("Fabric config path is not provided")

        return ProjectConfig(load_jobs_state(), load_fabric_config(), load_system_config(), load_configs_override())

    # best used when invoking from execution.
    @classmethod
    def from_conf_folder(cls, conf_folder):
        jobs_state = os.path.join(conf_folder, "jobs_state.yml")
        system_config = os.path.join(conf_folder, "system.yml")
        config_override = os.path.join(conf_folder, "config_override.yml")
        fabric_config = os.path.join(conf_folder, "fabrics.yml")

        return ProjectConfig.from_path(jobs_state, system_config, config_override, fabric_config)
